from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Foul


class LiveFoul(Foul, LiveEnhancedPbpItem):
    """
    Class for foul events
    """

    action_type = "foul"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        """
        returns the number of free throws resulting from the foul
        """
        if "1freethrow" in self.qualifiers:
            return 1
        elif "2freethrow" in self.qualifiers:
            return 2
        elif "3freethrow" in self.qualifiers:
            return 3

    @property
    def is_personal_foul(self):
        return self.sub_type == "personal" and not hasattr(self, "descriptor")

    @property
    def is_shooting_foul(self):
        return hasattr(self, "descriptor") and self.descriptor == "shooting"

    @property
    def is_loose_ball_foul(self):
        return hasattr(self, "descriptor") and self.stripped_descriptor == "looseball"

    @property
    def is_offensive_foul(self):
        return self.sub_type == "offensive" and (
            not hasattr(self, "descriptor") or self.descriptor != "charge"
        )

    @property
    def is_inbound_foul(self):
        return hasattr(self, "descriptor") and self.descriptor == "inbound"

    @property
    def is_away_from_play_foul(self):
        return (
            hasattr(self, "descriptor") and self.stripped_descriptor == "awayfromplay"
        )

    @property
    def is_clear_path_foul(self):
        return hasattr(self, "descriptor") and self.stripped_descriptor == "clearpath"

    @property
    def is_double_foul(self):
        return (
            hasattr(self, "descriptor")
            and self.descriptor == "double"
            and not self.is_technical
        )

    @property
    def is_technical(self):
        return self.sub_type == "technical" and not self.is_defensive_3_seconds

    @property
    def is_flagrant1(self):
        return (
            hasattr(self, "descriptor") and self.stripped_descriptor == "flagranttype1"
        )

    @property
    def is_flagrant2(self):
        return (
            hasattr(self, "descriptor") and self.stripped_descriptor == "flagranttype2"
        )

    @property
    def is_double_technical(self):
        return (
            self.sub_type == "technical"
            and hasattr(self, "descriptor")
            and self.descriptor == "double"
        )

    @property
    def is_defensive_3_seconds(self):
        return (
            hasattr(self, "descriptor")
            and self.stripped_descriptor == "defensive3second"
        )

    @property
    def is_delay_of_game(self):
        return self.stripped_sub_type == "delayofgame"

    @property
    def is_charge(self):
        return (
            self.sub_type == "offensive"
            and hasattr(self, "descriptor")
            and self.descriptor == "charge"
        )

    @property
    def is_personal_block_foul(self):
        return hasattr(self, "descriptor") and self.descriptor == "block"

    @property
    def is_personal_take_foul(self):
        return hasattr(self, "descriptor") and self.descriptor == "take"

    @property
    def is_shooting_block_foul(self):
        return hasattr(self, "descriptor") and self.descriptor == "block"



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import (
    FieldGoal,
    FreeThrow,
    JumpBall,
    Rebound,
    Turnover,
    Substitution,
    Timeout,
)
from pbpstats.resources.enhanced_pbp.rebound import EventOrderError


class StatsRebound(Rebound, StatsEnhancedPbpItem):
    """
    Class for rebound events
    """

    event_type = 4

    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        returns team id for team on offense for the shot that was rebounded
        """
        if self.is_real_rebound:
            return self.missed_shot.team_id
        if (
            isinstance(self.previous_event, Turnover)
            and not self.previous_event.is_no_turnover
        ):
            # shot clock turnover has place holder rebound after turnover
            # this correct team starts next possession
            team_ids = list(self.current_players.keys())
            return (
                team_ids[0]
                if team_ids[1] == self.previous_event.team_id
                else team_ids[1]
            )
        return self.previous_event.get_offense_team_id()

    @property
    def missed_shot(self):
        """
        returns :obj:`~pbpstats.resources.enhanced_pbp.field_goal.FieldGoal` or
        :obj:`~pbpstats.resources.enhanced_pbp.free_throw.FreeThrow` object
        for shot that was missed

        :raises: :obj:`~pbpstats.resources.enhanced_pbp.rebound.EventOrderError`:
            If rebound event is not immediately following a missed shot event.
        """
        if isinstance(self.previous_event, (FieldGoal, FreeThrow)):
            if not self.previous_event.is_made:
                return self.previous_event
        elif (
            isinstance(self.previous_event, Turnover)
            and self.previous_event.is_shot_clock_violation
        ):
            if isinstance(self.previous_event, FieldGoal):
                return self.previous_event.previous_event
        elif isinstance(self.previous_event, JumpBall):
            prev_event = self.previous_event.previous_event
            while isinstance(prev_event, (Substitution, Timeout)):
                prev_event = prev_event.previous_event
            if isinstance(prev_event, (FieldGoal, FreeThrow)):
                return prev_event
        raise EventOrderError(
            f"previous event: {self.previous_event} is not a missed free throw or field goal"
        )

    @property
    def is_placeholder(self):
        """
        returns True if rebound is a placeholder event, False otherwise.

        These are team rebounds on for example missed FT 1 of 2
        """
        return self.event_action_type != 0 and self.player1_id == 0

    @property
    def oreb(self):
        """
        returns True if rebound is an offensive rebound, False otherwise
        """
        return self.team_id == self.missed_shot.team_id



class StatsNbaLoaderBase(object):
    """
    Base Class for all stats.nba.com data loaders

    This class should not be instantiated directly
    """

    def make_list_of_dicts(self, results_set_index=0):
        """
        Creates list of dicts from source data

        :param int results_set_index: Index results are in. Default is 0
        :returns: list of dicts with data for results
        """
        headers = self.source_data["resultSets"][results_set_index]["headers"]
        rows = self.source_data["resultSets"][results_set_index]["rowSet"]
        deduped_rows = self.dedupe_events_row_set(rows)
        return [dict(zip(headers, row)) for row in deduped_rows]

    @staticmethod
    def dedupe_events_row_set(events_row_set):
        """
        Dedupes list of results while preserving order

        Used to dedupe events rowSets pbp response because some games have duplicate events

        :param list events_row_set: List of results from API Response
        :returns: deduped list of results
        """
        deduped_events_row_set = []
        for sublist in events_row_set:
            if sublist not in deduped_events_row_set:
                deduped_events_row_set.append(sublist)
        return deduped_events_row_set



"""
Instantiating a ``Game`` object will load all resources for the ``Game``
object that were set in the settings when the client was instantiated

The following code will instantiate the client and get possession data
for game id 0021900001 from files in ``/response_data`` subdirectories

.. code-block:: python

    from pbpstats.client import Client

    settings = {
        "dir": "/response_data",
        "Possessions": {"source": "file", "data_provider": "stats_nba"}
    }
    client = Client(settings)
    game = client.Game('0021900001')
    for possession in game.possessions.items:
        print(possession)
"""
import inspect

import pbpstats.client as client


class Game(object):
    """
    Class for loading resource data from data loaders with a ``parent_object`` of ``Game``

    :param str game_id: NBA Stats Game Id
    """

    def __init__(self, game_id):
        self.game_id = game_id
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        data_loaders = [
            a for a in attributes if a[0].endswith(client.DATA_LOADER_SUFFIX)
        ]
        data_source_map = {
            a[0].replace(client.DATA_SOURCE_SUFFIX, ""): a[1]
            for a in attributes
            if a[0].endswith(client.DATA_SOURCE_SUFFIX)
        }
        for data_loader in data_loaders:
            attr_name = data_loader[0].replace(client.DATA_LOADER_SUFFIX, "")
            source = data_source_map[attr_name]
            data = data_loader[1](game_id, source, self.data_directory)
            resource_cls = getattr(self, attr_name)
            setattr(
                self,
                client.PATTERN.sub("_", attr_name).lower(),
                resource_cls(data.items),
            )



import abc
import os
import json

from pbpstats import (
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
)
from pbpstats.overrides import IntDecoder
from pbpstats.resources.enhanced_pbp import (
    Ejection,
    EndOfPeriod,
    FieldGoal,
    Foul,
    FreeThrow,
    JumpBall,
    Substitution,
    Timeout,
    Turnover,
)


class InvalidNumberOfStartersException(Exception):
    """
    Class for exception when a team's 5 period starters can't be determined.

    You can add the correct period starters to
    overrides/missing_period_starters.json in your data directory to fix this.
    """

    pass


class StartOfPeriod(metaclass=abc.ABCMeta):
    """
    Class for start of period events
    """

    @abc.abstractclassmethod
    def get_period_starters(self, file_directory):
        """
        Gets player ids of players who started the period for each team

        :param str file_directory: directory in which overrides subdirectory exists
            containing period starter overrides when period starters can't be determined
            from parsing pbp events
        :returns: dict with list of player ids for each team
            with players on the floor at start of period
        :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
            If all 5 players that start the period for a team can't be determined.
        """
        pass

    @property
    def current_players(self):
        """
        returns period starters
        """
        return self.period_starters

    @property
    def league(self):
        """
        Returns League for game id.

        First 2 in game id represent league - 00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return G_LEAGUE_STRING
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    def get_team_starting_with_ball(self):
        """
        returns team id for team on starting period with the ball
        """
        if (self.period == 1 or self.period >= 5) and isinstance(
            self.next_event, JumpBall
        ):
            # period starts with jump ball - team that wins starts with the ball
            return self.next_event.team_id
        else:
            # find team id on first shot, non technical ft or turnover
            next_event = self.next_event
            while not (
                isinstance(next_event, (FieldGoal, Turnover))
                or (
                    isinstance(next_event, FreeThrow) and not next_event.is_technical_ft
                )
            ):
                next_event = next_event.next_event
            return next_event.team_id

    def get_offense_team_id(self):
        """
        returns team id for team on starting period on offense
        """
        return self.team_starting_with_ball

    def _get_players_who_started_period_with_team_map(self):
        starters = []
        subbed_in_players = []
        player_team_map = {}  # only player1 has team id in event, this is to track team
        event = self
        while event is not None and not isinstance(event, EndOfPeriod):
            if (
                not isinstance(event, Timeout)
                and event.player1_id != 0
                and hasattr(event, "team_id")
            ):
                player_id = event.player1_id
                if not isinstance(event, JumpBall):
                    # on jump balls team id is winning team, not guaranteed to be player1 team
                    player_team_map[player_id] = event.team_id
                if (
                    isinstance(event, Substitution)
                    and event.incoming_player_id is not None
                ):
                    player_team_map[event.incoming_player_id] = event.team_id
                    if (
                        event.incoming_player_id not in starters
                        and event.incoming_player_id not in subbed_in_players
                    ):
                        subbed_in_players.append(event.incoming_player_id)
                    if player_id not in starters and player_id not in subbed_in_players:
                        starters.append(player_id)

                is_technical_foul = isinstance(event, Foul) and (
                    event.is_technical or event.is_double_technical
                )
                if player_id not in starters and player_id not in subbed_in_players:
                    tech_ft_at_period_start = (
                        isinstance(event, FreeThrow) and event.clock == "12:00"
                    )
                    if not (
                        is_technical_foul
                        or isinstance(event, Ejection)
                        or tech_ft_at_period_start
                    ):
                        # ignore all techs because a player could get a technical foul when they aren't in the game
                        starters.append(player_id)
                # need player2_id and player3_id for players who play full period and never appear in an event as player_id - ex assists, blocks, steals, foul drawn
                if not isinstance(event, Substitution) and not (
                    is_technical_foul or isinstance(event, Ejection)
                ):
                    # ignore all techs because a player could get a technical foul when they aren't in the game
                    if (
                        hasattr(event, "player2_id")
                        and event.player2_id not in starters
                        and event.player2_id not in subbed_in_players
                    ):
                        starters.append(event.player2_id)
                    if (
                        hasattr(event, "player3_id")
                        and event.player3_id not in starters
                        and event.player3_id not in subbed_in_players
                    ):
                        starters.append(event.player3_id)
            event = event.next_event
        return starters, player_team_map

    def _split_up_starters_by_team(self, starters, player_team_map):
        starters_by_team = {}
        # for players who don't appear in event as player1 - won't be in player_team_map
        dangling_starters = []
        for player_id in starters:
            team_id = player_team_map.get(player_id)
            if team_id is not None:
                if team_id not in starters_by_team.keys():
                    starters_by_team[team_id] = []
                starters_by_team[team_id].append(player_id)
            else:
                dangling_starters.append(player_id)
        # if there is one dangling starter we can add it to team missing a starter
        if len(dangling_starters) == 1 and len(starters) == 10:
            for _, team_starters in starters_by_team.items():
                if len(team_starters) == 4:
                    team_starters += dangling_starters
        return starters_by_team

    def _check_both_teams_have_5_starters(self, starters_by_team, file_directory):
        """
        raises exception if either team does not have 5 starters
        """
        for team_id, starters in starters_by_team.items():
            if len(starters) != 5:
                # check if game and period are in overrides file
                if file_directory is None:
                    raise InvalidNumberOfStartersException(
                        f"GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}"
                    )

                missing_period_starters_file_path = (
                    f"{file_directory}/overrides/missing_period_starters.json"
                )
                if not os.path.isfile(missing_period_starters_file_path):
                    raise InvalidNumberOfStartersException(
                        f"GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}"
                    )

                with open(missing_period_starters_file_path) as f:
                    # hard code corrections for games with incorrect number of starters exceptions
                    missing_period_starters = json.loads(f.read(), cls=IntDecoder)
                game_id = (
                    self.game_id if self.league == NBA_STRING else int(self.game_id)
                )
                if (
                    self.game_id in missing_period_starters.keys()
                    and self.period in missing_period_starters[game_id].keys()
                    and team_id in missing_period_starters[game_id][self.period].keys()
                ):
                    starters_by_team[team_id] = missing_period_starters[self.game_id][
                        self.period
                    ][team_id]
                else:
                    raise InvalidNumberOfStartersException(
                        f"GameId: {game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}"
                    )

    def _get_period_starters_from_period_events(
        self, file_directory, ignore_missing_starters=False
    ):
        starters, player_team_map = self._get_players_who_started_period_with_team_map()

        starters_by_team = self._split_up_starters_by_team(starters, player_team_map)
        if not ignore_missing_starters:
            self._check_both_teams_have_5_starters(starters_by_team, file_directory)

        return starters_by_team

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats



class JumpBall(object):
    """
    Class for jump ball events
    """

    @property
    def winning_team(self):
        """
        returns team id that won the jump ball
        """
        return self.team_id

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats



"""
``DataNbaPossessionLoader`` loads possession data for a game and creates :obj:`~pbpstats.resources.possessions.possession.Possession` objects for each possession

The following code will load possession data for game id "0021900001" from a pbp file located in the ``/pbp`` subdirectory of the ``/data`` directory

.. code-block:: python

    from pbpstats.data_loader import DataNbaPossessionLoader

    possession_loader = DataNbaPossessionLoader("0021900001", "file", "/data")
    print(possession_loader.items[0].data)  # prints dict with the first possession of the game
"""
from pbpstats.data_loader.data_nba.enhanced_pbp_loader import DataNbaEnhancedPbpLoader
from pbpstats.data_loader.nba_possession_loader import NbaPossessionLoader
from pbpstats.resources.possessions.possession import Possession


class DataNbaPossessionLoader(NbaPossessionLoader):
    """
    Loads data.nba.com source possession data for game.
    Possessions are stored in items attribute as :obj:`~pbpstats.resources.possessions.possession.Possession` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `data_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "data_nba"
    resource = "Possessions"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):
        pbp_events = DataNbaEnhancedPbpLoader(game_id, source, file_directory)
        self.events = pbp_events.items
        events_by_possession = self._split_events_by_possession()
        self.items = [
            Possession(possession_events) for possession_events in events_by_possession
        ]
        self._add_extra_attrs_to_all_possessions()



OVERRIDES = {"personId": "player_id"}


class LiveBoxscoreItem(object):
    """
    Class for boxscore items from live data endpoint

    :param dict item: dict with boxscore stats from response
    :param int team_id: (optional) team id is not in dict with stats and can be added in here
    :param str team_abbreviation: (optional) team abbreviation is not in dict with stats and can be added in here
    """

    def __init__(self, item, team_id=None, team_abbreviation=None):
        self.team_id = team_id
        self.team_abbreviation = team_abbreviation
        for key, value in item.items():
            if key == "statistics":
                for stat_key, stat_value in value.items():
                    setattr(self, stat_key, stat_value)
            else:
                setattr(self, OVERRIDES.get(key, key), value)
        if hasattr(self, "firstName") and hasattr(self, "familyName"):
            self.name = f"{self.firstName} {self.familyName}"

    @property
    def data(self):
        """
        returns boxscore data dict
        """
        return self.__dict__

    @property
    def total_seconds(self):
        split = (
            self.minutes.replace("PT", "").replace("M", ":").replace("S", "").split(":")
        )
        return float(split[0]) * 60 + float(split[1])



import abc
import math

import pbpstats
from pbpstats.resources.enhanced_pbp import Foul, FreeThrow, Turnover, Violation


class FieldGoal(object):
    """
    Class for field goal events
    """

    @abc.abstractproperty
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        pass

    @abc.abstractproperty
    def shot_value(self):
        """
        returns 3 if shot is a 3 point attempt, 2 otherwise
        """
        pass

    @property
    def is_blocked(self):
        """
        returns True if shot was blocked, False otherwise
        """
        return not self.is_made and hasattr(self, "player3_id")

    @property
    def is_assisted(self):
        """
        returns True if shot was assisted, False otherwise
        """
        return self.is_made and hasattr(self, "player2_id")

    @property
    def rebound(self):
        """
        returns :obj:`~pbpstats.resources.enhanced_pbp.rebound.Rebound` item
        for the rebound of the shot, if it was missed, None otherwise
        """
        if not self.is_made and self.next_event.is_real_rebound:
            return self.next_event
        return None

    @property
    def is_heave(self):
        """
        returns True if shot was a last second heave, False otherwise
        """
        return (
            self.distance > pbpstats.HEAVE_DISTANCE_CUTOFF
            and self.seconds_remaining < pbpstats.HEAVE_TIME_CUTOFF
        )

    @property
    def is_corner_3(self):
        """
        returns True if shot was a corner 3, False otherwise
        """
        if self.shot_value != 3:
            return False
        if not hasattr(self, "locY") or self.locY is None:
            return False
        if self.locY <= 87:
            return True
        return False

    @property
    def distance(self):
        """
        returns shot distance in feet
        """
        if hasattr(self, "locX") and hasattr(self, "locY"):
            x_squared = self.locX ** 2
            y_squared = self.locY ** 2
            # unit for distance is off by factor of 10, divide by 10 to convert to feet
            shot_distance = math.sqrt(x_squared + y_squared) / 10
            return round(shot_distance, 1)
        else:
            # no coordinates - get shot distance from event description
            try:
                return int(self.description.split("'")[0].split(" ")[-1])
            except ValueError:
                return None

    @property
    def shot_type(self):
        """
        returns shot type string ('AtRim', 'ShortMidRange', 'LongMidRange', 'Arc3' or 'Corner3')
        """
        if self.shot_value == 3:
            if self.is_corner_3:
                return pbpstats.CORNER_3_STRING
            else:
                return pbpstats.ARC_3_STRING

        if self.distance is None:
            return pbpstats.UNKNOWN_SHOT_DISTANCE_STRING
        elif self.distance < pbpstats.AT_RIM_CUTOFF:
            return pbpstats.AT_RIM_STRING
        elif self.distance < pbpstats.SHORT_MID_RANGE_CUTOFF:
            return pbpstats.SHORT_MID_RANGE_STRING
        else:
            return pbpstats.LONG_MID_RANGE_STRING

    @property
    def is_putback(self):
        """
        returns True if shot is a 2pt attempt within 2 seconds of an
        offensive rebound attempted by the same player who got the rebound
        """
        if self.is_assisted or self.shot_value == 3:
            return False
        prev_evt = self.previous_event
        if prev_evt is None:
            return False
        prev_evt_is_shooting_foul = isinstance(prev_evt, Foul) and (
            prev_evt.is_shooting_foul or prev_evt.is_shooting_block_foul
        )
        prev_evt_is_goaltend = (
            isinstance(prev_evt, Violation) and prev_evt.is_goaltend_violation
        )
        if (
            prev_evt_is_shooting_foul or prev_evt_is_goaltend
        ) and self.clock == prev_evt.clock:
            # sometimes foul event is between rebound and shot on an and 1 or goaltend is between rebound and made shot event
            prev_evt = prev_evt.previous_event
            if prev_evt is None:
                return False
        if not hasattr(prev_evt, "is_real_rebound"):
            return False
        if not prev_evt.is_real_rebound:
            return False
        return (
            prev_evt.oreb
            and prev_evt.player1_id == self.player1_id
            and prev_evt.seconds_remaining - self.seconds_remaining <= 2
        )

    @property
    def is_and1(self):
        """
        returns True if shot was an and 1, False otherwise
        """
        if self.is_make_that_does_not_end_possession:
            free_throws = [
                event
                for event in self.get_all_events_at_current_time()
                if isinstance(event, FreeThrow)
            ]
            for ft in free_throws:
                if "And 1" in ft.free_throw_type and ft.player1_id == self.player1_id:
                    return True
        return False

    @property
    def shot_data(self):
        """
        returns a dict with detailed shot data
        """
        team_ids = list(self.current_players.keys())
        opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
        shot_data = {
            "PlayerId": self.player1_id,
            "TeamId": self.team_id,
            "OpponentTeamId": opponent_team_id,
            "LineupId": self.lineup_ids[self.team_id],
            "OpponentLineupId": self.lineup_ids[opponent_team_id],
            "Made": self.is_made,
            "X": self.locX if hasattr(self, "locX") else None,
            "Y": self.locY if hasattr(self, "locY") else None,
            "Time": self.seconds_remaining,
            "ShotValue": self.shot_value,
            "Putback": self.is_putback,
            "ShotType": self.shot_type,
            "ScoreMargin": self.score_margin,
            "EventNum": self.event_num,
            "IsAnd1": self.is_and1,
        }
        if self.is_made:
            shot_data["Assisted"] = self.is_assisted
            if self.is_assisted:
                shot_data["AssistPlayerId"] = self.player2_id
        if not self.is_made:
            shot_data["Blocked"] = self.is_blocked
            if self.is_blocked:
                shot_data["BlockPlayerId"] = self.player3_id
        if self.is_second_chance_event():
            prev_event = self.previous_event
            while not (
                hasattr(prev_event, "is_real_rebound") and prev_event.is_real_rebound
            ):
                prev_event = prev_event.previous_event
            shot_data["SecondsSinceOReb"] = (
                prev_event.seconds_remaining - self.seconds_remaining
            )
            shot_data["OrebShotPlayerId"] = prev_event.missed_shot.player1_id
            shot_data["OrebReboundPlayerId"] = prev_event.player1_id
            if prev_event.player1_id != 0:
                rebound_shot_type = prev_event.missed_shot.shot_type
                if (
                    isinstance(prev_event.missed_shot, FieldGoal)
                    and prev_event.missed_shot.is_blocked
                ):
                    rebound_shot_type += pbpstats.BLOCKED_STRING
            elif (
                isinstance(prev_event.missed_shot, FieldGoal)
                and prev_event.missed_shot.is_blocked
            ):
                # separate blocked from non blocked because blocked won't have shot clock reset
                rebound_shot_type = "TeamBlocked"
            else:
                rebound_shot_type = "Team"
            shot_data["OrebShotType"] = rebound_shot_type
        return shot_data

    def _check_if_make_ends_possession_when_1_foul(
        self, foul_event, events_at_shot_time
    ):
        """
        helper method for determining if made shot ends possession
        when there was 1 foul at the time of the shot
        """
        foul_team_team_id = foul_event.team_id
        shooter_team_id = self.team_id
        if foul_event.is_flagrant:
            # flagrant foul and 1
            if foul_team_team_id != shooter_team_id:
                return True

        other_makes_at_time_of_shot = []
        for event in events_at_shot_time:
            if (
                isinstance(event, FieldGoal)
                and event.event_num != self.event_num
                and event.is_made
                and self.team_id == event.team_id
            ):
                other_makes_at_time_of_shot.append(event)
        if shooter_team_id != foul_team_team_id:
            # check FT 1 of 1s at time of shot
            ft_1_of_1s_at_time_of_shot = []
            for event in events_at_shot_time:
                if (
                    isinstance(event, FreeThrow)
                    and (event.is_ft_1_of_1 or event.is_ft_1pt)
                    and not event.is_technical_ft
                ):
                    ft_1_of_1s_at_time_of_shot.append(event)

            if (
                len(other_makes_at_time_of_shot) == 1
                and len(ft_1_of_1s_at_time_of_shot) == 1
            ):
                if (
                    ft_1_of_1s_at_time_of_shot[0].player1_id
                    == other_makes_at_time_of_shot[0].player1_id
                ):
                    return False
                elif ft_1_of_1s_at_time_of_shot[0].player1_id == self.player1_id:
                    return True
            elif len(ft_1_of_1s_at_time_of_shot) > 0:
                for ft_event in ft_1_of_1s_at_time_of_shot:
                    if ft_event.team_id == shooter_team_id:
                        return True
            else:
                # no free throws - check for lane violation and offensive goaltending
                for event in events_at_shot_time:
                    if isinstance(event, Turnover) and (
                        event.is_lane_violation or event.is_offensive_goaltending
                    ):
                        return True
                    if isinstance(event, Violation) and event.is_double_lane_violation:
                        return True
        return False

    def _check_if_make_ends_possession_when_not_1_foul(
        self, events_at_shot_time, fouls_at_time_of_shot
    ):
        """
        helper method for determining if made shot ends possession
        when there was not 1 foul at the time of the shot
        """
        shooter_team_id = self.team_id
        if shooter_team_id not in [event.team_id for event in fouls_at_time_of_shot]:
            ft_1_of_1s_at_time_of_shot = []
            for event in events_at_shot_time:
                if (
                    isinstance(event, FreeThrow)
                    and (event.is_ft_1_of_1 or event.is_ft_1pt)
                    and not event.is_technical_ft
                ):
                    ft_1_of_1s_at_time_of_shot.append(event)

            if len(ft_1_of_1s_at_time_of_shot) == 1:
                if ft_1_of_1s_at_time_of_shot[0].team_id == shooter_team_id:
                    return True
            elif len(ft_1_of_1s_at_time_of_shot) > 1:
                for ft_event in ft_1_of_1s_at_time_of_shot:
                    if ft_event.player1_id == self.player1_id:
                        return True
        else:
            opponent_fouls = [
                event
                for event in fouls_at_time_of_shot
                if event.team_id != shooter_team_id
            ]
            if 1 in [event.number_of_fta_for_foul for event in opponent_fouls]:
                # rarely gets here but it is when there is an and1 and then another foul at the same time
                # for example a loose ball foul going for the rebound on the FT
                return True
        return False

    @property
    def is_make_that_does_not_end_possession(self):
        """
        returns True if shot is a made shot that does not end the possession due to a foul, False otherwise
        """
        # check for foul at time of shot
        fouls_at_time_of_shot = []
        # ignore technical fouls and delay of game fouls when getting last foul
        events_at_shot_time = self.get_all_events_at_current_time()
        for event in events_at_shot_time:
            if (
                isinstance(event, Foul)
                and not event.is_delay_of_game
                and not event.is_technical
            ):
                fouls_at_time_of_shot.append(event)
        if len(fouls_at_time_of_shot) == 1:
            foul = fouls_at_time_of_shot[0]
            return self._check_if_make_ends_possession_when_1_foul(
                foul, events_at_shot_time
            )
        return self._check_if_make_ends_possession_when_not_1_foul(
            events_at_shot_time, fouls_at_time_of_shot
        )

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        if self.distance is not None:
            stats += self._get_shot_distance_stat_items()
            if self.is_heave:
                stats.append(self._get_heave_stat_item())

        team_ids = list(self.current_players.keys())
        opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
        if self.is_made and not self.is_assisted:
            stats += self._get_unassisted_stat_items(
                is_second_chance_event, is_penalty_event
            )
        elif self.is_assisted:
            stats += self._get_assisted_stat_items(
                is_second_chance_event, is_penalty_event
            )
        elif self.is_blocked:
            stats += self._get_blocked_stat_items(
                is_second_chance_event, is_penalty_event, opponent_team_id
            )
        else:
            stats += self._get_missed_stat_items(
                is_second_chance_event, is_penalty_event
            )

        if self.is_made:
            # add plus minus and opponent points - used for lineup/wowy stats to get net rating
            for team_id, players in self.current_players.items():
                multiplier = 1 if team_id == self.team_id else -1
                for player_id in players:
                    stat_item = {
                        "player_id": player_id,
                        "team_id": team_id,
                        "stat_key": pbpstats.PLUS_MINUS_STRING,
                        "stat_value": self.shot_value * multiplier,
                    }
                    stats.append(stat_item)
                    if multiplier == -1:
                        opponent_points_stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "stat_key": pbpstats.OPPONENT_POINTS,
                            "stat_value": self.shot_value,
                        }
                        stats.append(opponent_points_stat_item)

        lineups_ids = self.lineup_ids
        for stat in stats:
            opponent_team_id = (
                team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
            )
            stat["lineup_id"] = lineups_ids[stat["team_id"]]
            stat["opponent_team_id"] = opponent_team_id
            stat["opponent_lineup_id"] = lineups_ids[opponent_team_id]

        return self.base_stats + stats

    def _get_shot_distance_stat_items(self):
        stats = []
        if self.shot_value == 2:
            distance_stat_key = pbpstats.TOTAL_2PT_SHOT_DISTANCE_STRING
            count_stat_key = pbpstats.TOTAL_2PT_SHOTS_WITH_DISTANCE
        else:
            distance_stat_key = pbpstats.TOTAL_3PT_SHOT_DISTANCE_STRING
            count_stat_key = pbpstats.TOTAL_3PT_SHOTS_WITH_DISTANCE
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": distance_stat_key,
                "stat_value": self.distance,
            }
        )
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": count_stat_key,
                "stat_value": 1,
            }
        )
        return stats

    def _get_heave_stat_item(self):
        if self.is_made:
            return {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": pbpstats.HEAVE_MAKES_STRING,
                "stat_value": 1,
            }
        else:
            return {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": pbpstats.HEAVE_MISSES_STRING,
                "stat_value": 1,
            }

    def _get_unassisted_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        stat_key = f"{pbpstats.UNASSISTED_STRING}{self.shot_type}"
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": stat_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_stat_key = f"{pbpstats.SECOND_CHANCE_STRING}{stat_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_stat_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_stat_key = f"{pbpstats.PENALTY_STRING}{stat_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_stat_key,
                    "stat_value": 1,
                }
            )
        if self.is_putback:
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": pbpstats.PUTBACKS_STRING,
                    "stat_value": 1,
                }
            )
        return stats

    def _get_assisted_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        scorer_key = f"{pbpstats.ASSISTED_STRING}{self.shot_type}"
        assist_key = f"{self.shot_type}{pbpstats.ASSISTS_STRING}"
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": scorer_key,
                "stat_value": 1,
            }
        )
        stats.append(
            {
                "player_id": self.player2_id,
                "team_id": self.team_id,
                "stat_key": assist_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_scorer_key = f"{pbpstats.SECOND_CHANCE_STRING}{scorer_key}"
            second_chance_assist_key = f"{pbpstats.SECOND_CHANCE_STRING}{assist_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_scorer_key,
                    "stat_value": 1,
                }
            )
            stats.append(
                {
                    "player_id": self.player2_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_assist_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_scorer_key = f"{pbpstats.PENALTY_STRING}{scorer_key}"
            penalty_assist_key = f"{pbpstats.PENALTY_STRING}{assist_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_scorer_key,
                    "stat_value": 1,
                }
            )
            stats.append(
                {
                    "player_id": self.player2_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_assist_key,
                    "stat_value": 1,
                }
            )
        assist_to_key = (
            f"{self.player2_id}:AssistsTo:{self.player1_id}:{self.shot_type}"
        )
        stats.append(
            {
                "player_id": self.player2_id,
                "team_id": self.team_id,
                "stat_key": assist_to_key,
                "stat_value": 1,
            }
        )
        return stats

    def _get_blocked_stat_items(
        self, is_second_chance_event, is_penalty_event, opponent_team_id
    ):
        stats = []
        shot_key = f"{self.shot_type}{pbpstats.BLOCKED_STRING}"
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": shot_key,
                "stat_value": 1,
            }
        )
        block_key = f"{pbpstats.BLOCKED_STRING}{self.shot_type}"
        stats.append(
            {
                "player_id": self.player3_id,
                "team_id": opponent_team_id,
                "stat_key": block_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_shot_key = f"{pbpstats.SECOND_CHANCE_STRING}{shot_key}"
            second_chance_block_key = f"{pbpstats.SECOND_CHANCE_STRING}{block_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_shot_key,
                    "stat_value": 1,
                }
            )
            stats.append(
                {
                    "player_id": self.player3_id,
                    "team_id": opponent_team_id,
                    "stat_key": second_chance_block_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_shot_key = f"{pbpstats.PENALTY_STRING}{shot_key}"
            penalty_block_key = f"{pbpstats.PENALTY_STRING}{block_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_shot_key,
                    "stat_value": 1,
                }
            )
            stats.append(
                {
                    "player_id": self.player3_id,
                    "team_id": opponent_team_id,
                    "stat_key": penalty_block_key,
                    "stat_value": 1,
                }
            )
        return stats

    def _get_missed_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        stat_key = f"{pbpstats.MISSED_STRING}{self.shot_type}"
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": stat_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_stat_key = f"{pbpstats.SECOND_CHANCE_STRING}{stat_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_stat_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_stat_key = f"{pbpstats.PENALTY_STRING}{stat_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_stat_key,
                    "stat_value": 1,
                }
            )
        return stats



"""
The ``Games`` class has some basic properties for handling game data
"""
from pbpstats.resources.base import Base


class Games(Base):
    """
    Class for games items

    :param list items: list of either
        :obj:`~pbpstats.resources.games.stats_nba_game_item.StatsNbaGameItem` or
        :obj:`~pbpstats.resources.games.data_nba_game_item.DataNbaGameItem` items,
        typically from a games data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns list of dicts with game items
        """
        return [item.data for item in self.items]

    @property
    def final_games(self):
        """
        returns list of dicts with final game items
        """
        return [item.data for item in self.items if item.is_final]



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import FieldGoal


class LiveFieldGoal(FieldGoal, LiveEnhancedPbpItem):
    """
    Class for field goal events
    """

    action_type = ["2pt", "3pt"]

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def shot_value(self):
        """
        returns 3 if shot is a 3 point attempt, 2 otherwise
        """
        return 3 if self.action_type == "3pt" else 2

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return self.shot_result == "Made"



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Turnover


class StatsTurnover(Turnover, StatsEnhancedPbpItem):
    """
    Class for Turnover events
    """

    event_type = 5

    def __init__(self, *args):
        super().__init__(*args)

    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        if self.is_no_turnover:
            return self.previous_event.get_offense_team_id()
        return self.team_id

    @property
    def is_no_turnover(self):
        return self.event_action_type == 0

    @property
    def is_bad_pass(self):
        return self.event_action_type == 1 and self.is_steal

    @property
    def is_lost_ball(self):
        return self.event_action_type == 2 and self.is_steal

    @property
    def is_travel(self):
        return self.event_action_type == 4

    @property
    def is_3_second_violation(self):
        return self.event_action_type == 8

    @property
    def is_shot_clock_violation(self):
        return self.event_action_type == 11

    @property
    def is_offensive_goaltending(self):
        return self.event_action_type == 15

    @property
    def is_lane_violation(self):
        return self.event_action_type == 17

    @property
    def is_kicked_ball(self):
        return self.event_action_type == 19

    @property
    def is_step_out_of_bounds(self):
        return self.event_action_type == 39

    @property
    def is_lost_ball_out_of_bounds(self):
        # some labelled as lost ball but should be lost ball out of bounds (missing player3 id)
        return self.event_action_type == 40 or (
            self.event_action_type == 2 and not self.is_steal
        )

    @property
    def is_bad_pass_out_of_bounds(self):
        # some labelled as bad pass but should be bad pass out of bounds (missing player3 id)
        return self.event_action_type == 45 or (
            self.event_action_type == 1 and not self.is_steal
        )



from pbpstats.resources.enhanced_pbp.stats_nba.ejection import StatsEjection
from pbpstats.resources.enhanced_pbp.stats_nba.end_of_period import StatsEndOfPeriod
from pbpstats.resources.enhanced_pbp.stats_nba.field_goal import StatsFieldGoal
from pbpstats.resources.enhanced_pbp.stats_nba.foul import StatsFoul
from pbpstats.resources.enhanced_pbp.stats_nba.free_throw import StatsFreeThrow
from pbpstats.resources.enhanced_pbp.stats_nba.jump_ball import StatsJumpBall
from pbpstats.resources.enhanced_pbp.stats_nba.rebound import StatsRebound
from pbpstats.resources.enhanced_pbp.stats_nba.replay import StatsReplay
from pbpstats.resources.enhanced_pbp.stats_nba.start_of_period import StatsStartOfPeriod
from pbpstats.resources.enhanced_pbp.stats_nba.substitution import StatsSubstitution
from pbpstats.resources.enhanced_pbp.stats_nba.timeout import StatsTimeout
from pbpstats.resources.enhanced_pbp.stats_nba.turnover import StatsTurnover
from pbpstats.resources.enhanced_pbp.stats_nba.violation import StatsViolation

__all__ = [
    "StatsEjection",
    "StatsEndOfPeriod",
    "StatsFieldGoal",
    "StatsFoul",
    "StatsFreeThrow",
    "StatsJumpBall",
    "StatsRebound",
    "StatsReplay",
    "StatsStartOfPeriod",
    "StatsSubstitution",
    "StatsTimeout",
    "StatsTurnover",
    "StatsViolation",
]



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Replay


class StatsReplay(Replay, StatsEnhancedPbpItem):
    """
    Class for replay events
    """

    event_type = 18

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def support_ruling(self):
        return self.event_action_type == 4

    @property
    def overturn_ruling(self):
        return self.event_action_type == 5

    @property
    def ruling_stands(self):
        return self.event_action_type == 6



"""
``LiveEnhancedPbpLoader`` loads pbp data for a game and creates :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects for each event

Enhanced data for each event includes current players on floor, score, fouls to give and number of fouls committed by each player,
plus additional data depending on event type

The following code will load pbp data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import LiveEnhancedPbpLoader

    pbp_loader = LiveEnhancedPbpLoader("0021900001", "file", "/data")
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.live.pbp_loader import LivePbpLoader
from pbpstats.data_loader.nba_enhanced_pbp_loader import NbaEnhancedPbpLoader
from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_factory import (
    LiveEnhancedPbpFactory,
)
from pbpstats.resources.enhanced_pbp import Rebound


class LiveEnhancedPbpLoader(LivePbpLoader, NbaEnhancedPbpLoader):
    """
    Loads data.nba.com source enhanced pbp data for game.
    Events are stored in items attribute as :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `live_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
        If all 5 players that start the period for a team can't be determined.
        You can add the correct period starters to overrides/missing_period_starters.json in your data directory to fix this.
    """

    data_provider = "live"
    resource = "EnhancedPbp"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):
        super().__init__(game_id, source, file_directory)

    def _make_pbp_items(self):
        factory = LiveEnhancedPbpFactory()
        self.items = [
            factory.get_event_class(event["actionType"], event.get("subType", ""))(
                event, self.game_id
            )
            for event in self.data
        ]
        self._add_extra_attrs_to_all_events()
        self._change_team_id_on_drebs()

    def _change_team_id_on_drebs(self):
        """
        live pbp changes possession on dreb, swap team to be consistent with other sources
        """
        for event in self.items:
            if isinstance(event, Rebound):
                if event.is_real_rebound and not event.oreb:
                    event.offense_team_id = event.previous_event.offense_team_id



import abc


class Base(metaclass=abc.ABCMeta):
    """
    base class for all resources classes
    all resource classes should inherit from this
    """

    @abc.abstractproperty
    def data(self):
        pass



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Violation


class StatsViolation(Violation, StatsEnhancedPbpItem):
    """
    Class for violation events
    """

    event_type = 7

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 1

    @property
    def is_goaltend_violation(self):
        return self.event_action_type == 2

    @property
    def is_lane_violation(self):
        return self.event_action_type == 3

    @property
    def is_jumpball_violation(self):
        return self.event_action_type == 4

    @property
    def is_kicked_ball_violation(self):
        return self.event_action_type == 5

    @property
    def is_double_lane_violation(self):
        return self.event_action_type == 6



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Timeout


class LiveTimeout(Timeout, LiveEnhancedPbpItem):
    """
    Class for timeout events
    """

    action_type = "timeout"

    def __init__(self, *args):
        super().__init__(*args)



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import FreeThrow


class StatsFreeThrow(FreeThrow, StatsEnhancedPbpItem):
    """
    Class for free throw events
    """

    event_type = 3

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return "MISS " not in self.description

    def get_offense_team_id(self):
        """
        returns team id that took the shot
        """
        return self.team_id

    @property
    def is_ft_1_of_1(self):
        return self.event_action_type == 10

    @property
    def is_ft_1_of_2(self):
        return self.event_action_type == 11

    @property
    def is_ft_2_of_2(self):
        return self.event_action_type == 12

    @property
    def is_ft_1_of_3(self):
        return self.event_action_type == 13

    @property
    def is_ft_2_of_3(self):
        return self.event_action_type == 14

    @property
    def is_ft_3_of_3(self):
        return self.event_action_type == 15

    @property
    def is_technical_ft(self):
        return " Technical" in self.description

    @property
    def is_ft_1pt(self):
        """
        returns True if free throw is a 1 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        return self.event_action_type == 30 or self.event_action_type == 35

    @property
    def is_ft_2pt(self):
        """
        returns True if free throw is a 2 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        return self.event_action_type == 31 or self.event_action_type == 36

    @property
    def is_ft_3pt(self):
        """
        returns True if free throw is a 3 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        return self.event_action_type == 32 or self.event_action_type == 37



"""
``LiveBoxscoreLoader`` loads boxscore data for a game and creates :obj:`~pbpstats.resources.boxscore.live_boxscore_item.LiveBoxscoreItem` objects for each player and team

The following code will load boxscore data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import LiveBoxscoreLoader

    boxscore_loader = LiveBoxscoreLoader("0021900001", "file", "/data")
    print(boxscore_loader.items[0].data) # prints dict with a player's boxscore data for game
"""

import json
import os

from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.live.file_loader import LiveFileLoader
from pbpstats.data_loader.live.web_loader import LiveWebLoader
from pbpstats.resources.boxscore.live_boxscore_item import LiveBoxscoreItem


class LiveBoxscoreLoader(LiveFileLoader, LiveWebLoader):
    """
    Loads data.nba.com source boxscore data for game.
    Team/Player data is stored in items attribute as :obj:`~pbpstats.resources.boxscore.live_boxscore_item.LiveBoxscoreItem` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `live_<game_id>.json` in the `/game_details` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "live"
    resource = "Boxscore"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):
        self.game_id = game_id
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_boxscore_items()

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.file_path = f"{self.file_directory}/game_details/live_{self.game_id}.json"
        self._load_data_from_file()

    def _from_web(self):
        league = self.league.upper()
        self.url = f"https://nba-prod-us-east-1-mediaops-stats.s3.amazonaws.com/{league}/liveData/boxscore/boxscore_{self.game_id}.json"
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f"{self.file_directory}/game_details/live_{self.game_id}.json"
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    def _make_boxscore_items(self):
        """
        makes :obj:`~pbpstats.resources.boxscore.LiveBoxscoreItem` items for each player/team
        """
        home = self.data["homeTeam"]
        away = self.data["awayTeam"]
        self.items = [
            LiveBoxscoreItem(
                item, team_id=away["teamId"], team_abbreviation=away["teamTricode"]
            )
            for item in away["players"]
        ]
        self.items += [
            LiveBoxscoreItem(
                item, team_id=home["teamId"], team_abbreviation=home["teamTricode"]
            )
            for item in home["players"]
        ]
        self.items.append(
            LiveBoxscoreItem(
                away["statistics"],
                team_id=away["teamId"],
                team_abbreviation=away["teamTricode"],
            )
        )
        self.items.append(
            LiveBoxscoreItem(
                home["statistics"],
                team_id=home["teamId"],
                team_abbreviation=home["teamTricode"],
            )
        )

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["game"]



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import StartOfPeriod


class DataStartOfPeriod(StartOfPeriod, DataEnhancedPbpItem):
    """
    Class for start of period events
    """

    event_type = 12

    def __init__(self, *args):
        super().__init__(*args)

    def get_period_starters(self, file_directory=None):
        """
        Gets player ids of players who started the period for each team

        :param str file_directory: directory in which overrides subdirectory exists
            containing period starter overrides when period starters can't be determined
            from parsing pbp events
        :returns: dict with list of player ids for each team
            with players on the floor at start of period
        :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
            If all 5 players that start the period for a team can't be determined.
        """
        return self._get_period_starters_from_period_events(file_directory)



"""
Instantiating a ``Client`` object will load data loader objects
for resources specified in settings dict.

The following code will instantiate the client and get Possession data
for game id 0021900001 from files in ``/response_data`` subdirectories

.. code-block:: python

    from pbpstats.client import Client

    settings = {
        "dir": "/response_data",
        "Possessions": {"source": "file", "data_provider": "stats_nba"}
    }
    client = Client(settings)
    game = client.Game('0021900001')
    for possession in game.possessions.items:
        print(possession)
"""
import re

import pbpstats.objects as objects
import pbpstats.resources as resources
from pbpstats.data_loader.factory import DataLoaderFactory


DATA_LOADER_SUFFIX = "DataLoaderClass"
DATA_SOURCE_SUFFIX = "DataSource"
PATTERN = re.compile(r"(?<!^)(?=[A-Z])")  # for converting camel case to snake case


class Client(object):
    """
    :param dict settings: Dict with data that specifies which data loaders should be used.
        ``dir`` key is optional, but recommended and should point to the directory you have set up
        that either already contains response data or where you want to store the response data.
        Other keys in the settings dict should be resources from the :py:mod:`~pbpstats.resources` module
        and their values should be a dict with ``source`` ('file' or 'web') and ``data_provider`` ('stats_nba' or 'data_nba' or 'live')
    """

    def __init__(self, settings):
        data_loader = DataLoaderFactory()
        self.settings = settings
        self.data_directory = settings.get("dir")
        self._load_objects()
        self._load_resources()
        for resource, value in settings.items():
            if resource in data_loader.loaders.keys():
                resource_data_loaders = data_loader.get_data_loader(
                    value["data_provider"], resource
                )
                for resource_data_loader in resource_data_loaders:
                    parent_cls_name = resource_data_loader.parent_object
                    setattr(
                        getattr(self, parent_cls_name),
                        f"{resource}{DATA_LOADER_SUFFIX}",
                        resource_data_loader,
                    )
                    setattr(
                        getattr(self, parent_cls_name),
                        f"{resource}{DATA_SOURCE_SUFFIX}",
                        value["source"],
                    )
                    setattr(
                        getattr(self, parent_cls_name),
                        resource,
                        self.resource_dict[resource],
                    )

    def _load_objects(self):
        """
        loads classes from objects package
        """
        object_dict = dict(
            [
                (name, cls)
                for name, cls in objects.__dict__.items()
                if isinstance(cls, type)
            ]
        )
        for name, object_cls in object_dict.items():
            setattr(self, name, object_cls)
            setattr(getattr(self, name), "data_directory", self.data_directory)

    def _load_resources(self):
        """
        loads classes from resources package
        """
        self.resource_dict = dict(
            [
                (name, cls)
                for name, cls in resources.__dict__.items()
                if isinstance(cls, type)
            ]
        )



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Foul


class DataFoul(Foul, DataEnhancedPbpItem):
    """
    Class for foul events
    """

    event_type = 6

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def number_of_fta_for_foul(self):
        """
        returns the number of free throws resulting from the foul
        """
        if "(1 FTA)" in self.description:
            return 1
        elif "(2 FTA)" in self.description:
            return 2
        elif "(3 FTA)" in self.description:
            return 3

    @property
    def is_personal_foul(self):
        return self.event_action_type == 1

    @property
    def is_shooting_foul(self):
        return self.event_action_type == 2

    @property
    def is_loose_ball_foul(self):
        return self.event_action_type == 3

    @property
    def is_offensive_foul(self):
        return self.event_action_type == 4

    @property
    def is_inbound_foul(self):
        return self.event_action_type == 5

    @property
    def is_away_from_play_foul(self):
        return self.event_action_type == 6

    @property
    def is_clear_path_foul(self):
        return self.event_action_type == 9

    @property
    def is_double_foul(self):
        return self.event_action_type == 10

    @property
    def is_technical(self):
        return self.event_action_type in [11, 12, 13, 18, 19, 25, 30]

    @property
    def is_flagrant1(self):
        return self.event_action_type == 14

    @property
    def is_flagrant2(self):
        return self.event_action_type == 15

    @property
    def is_double_technical(self):
        return self.event_action_type == 16

    @property
    def is_defensive_3_seconds(self):
        return self.event_action_type == 17

    @property
    def is_delay_of_game(self):
        return self.event_action_type == 18

    @property
    def is_charge(self):
        return self.event_action_type == 26

    @property
    def is_personal_block_foul(self):
        return self.event_action_type == 27

    @property
    def is_personal_take_foul(self):
        return self.event_action_type == 28

    @property
    def is_shooting_block_foul(self):
        return self.event_action_type == 29



import requests

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    PLAYOFFS_STRING,
    REGULAR_SEASON_STRING,
    PLAY_IN_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.data_loader.abs_data_loader import AbsDataLoader
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase


class StatsNbaWebLoader(AbsDataLoader, StatsNbaLoaderBase):
    """
    Base class for loading data from data.nba.com API request.

    All stats.nba.com data loader classes should inherit from this class.

    This class should not be instantiated directly.
    """

    def _load_request_data(self):
        response = requests.get(
            self.base_url, self.parameters, headers=HEADERS, timeout=REQUEST_TIMEOUT
        )
        if response.status_code == 200:
            self.source_data = response.json()
            self._save_data_to_file()
        else:
            response.raise_for_status()

    @property
    def data(self):
        """
        returns data from response JSON as a list of dicts
        """
        return self.make_list_of_dicts()

    @property
    def league(self):
        """
        Returns League for game id.

        First 2 in game id represent league - 00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return G_LEAGUE_STRING
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    @property
    def season(self):
        """
        Returns season for game id

        4th and 5th characters in game id represent season year
        ex. for 2016-17 season 4th and 5th characters would be 16 and season should return 2016-17
        For WNBA just returns season year
        """
        digit4 = int(self.game_id[3])
        digit5 = int(self.game_id[4])
        if digit4 == 9:
            if digit5 == 9:
                return "1999" if self.league == WNBA_STRING else "1999-00"
            else:
                return (
                    f"19{digit4}{digit5}"
                    if self.league == WNBA_STRING
                    else f"19{digit4}{digit5}-{digit4}{digit5 + 1}"
                )
        elif digit5 == 9:
            return (
                f"20{digit4}{digit5}"
                if self.league == WNBA_STRING
                else f"20{digit4}{digit5}-{digit4 + 1}0"
            )
        else:
            return (
                f"20{digit4}{digit5}"
                if self.league == WNBA_STRING
                else f"20{digit4}{digit5}-{digit4}{digit5 + 1}"
            )

    @property
    def season_type(self):
        """
        Returns season type for game id

        3rd character in game id represent season type - 2 for reg season, 4 for playoffs, 5 for play in
        """
        if self.game_id[2] == "4":
            return PLAYOFFS_STRING
        elif self.game_id[2] == "2":
            return REGULAR_SEASON_STRING
        elif self.game_id[2] == "5":
            return PLAY_IN_STRING



from pbpstats.data_loader.data_nba.boxscore_loader import DataNbaBoxscoreLoader
from pbpstats.data_loader.data_nba.enhanced_pbp_loader import DataNbaEnhancedPbpLoader
from pbpstats.data_loader.data_nba.pbp_loader import DataNbaPbpLoader
from pbpstats.data_loader.data_nba.possessions_loader import DataNbaPossessionLoader
from pbpstats.data_loader.data_nba.schedule_loader import DataNbaScheduleLoader
from pbpstats.data_loader.live.boxscore_loader import LiveBoxscoreLoader
from pbpstats.data_loader.live.enhanced_pbp_loader import LiveEnhancedPbpLoader
from pbpstats.data_loader.live.pbp_loader import LivePbpLoader
from pbpstats.data_loader.live.possessions_loader import LivePossessionLoader
from pbpstats.data_loader.stats_nba.boxscore_loader import StatsNbaBoxscoreLoader
from pbpstats.data_loader.stats_nba.enhanced_pbp_loader import StatsNbaEnhancedPbpLoader
from pbpstats.data_loader.stats_nba.game_finder_loader import StatsNbaGameFinderLoader
from pbpstats.data_loader.stats_nba.pbp_loader import StatsNbaPbpLoader
from pbpstats.data_loader.stats_nba.possessions_loader import StatsNbaPossessionLoader
from pbpstats.data_loader.stats_nba.scoreboard_loader import StatsNbaScoreboardLoader
from pbpstats.data_loader.stats_nba.shots_loader import StatsNbaShotsLoader

__all__ = [
    "DataNbaBoxscoreLoader",
    "DataNbaEnhancedPbpLoader",
    "DataNbaPbpLoader",
    "DataNbaPossessionLoader",
    "DataNbaScheduleLoader",
    "LiveBoxscoreLoader",
    "LiveEnhancedPbpLoader",
    "LivePbpLoader",
    "LivePossessionLoader",
    "StatsNbaBoxscoreLoader",
    "StatsNbaEnhancedPbpLoader",
    "StatsNbaGameFinderLoader",
    "StatsNbaPbpLoader",
    "StatsNbaPossessionLoader",
    "StatsNbaScoreboardLoader",
    "StatsNbaShotsLoader",
]



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import FieldGoal


class DataFieldGoal(FieldGoal, DataEnhancedPbpItem):
    """
    Class for field goal events
    """

    event_type = [1, 2]

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return self.event_type == 1

    @property
    def shot_value(self):
        """
        returns 3 if shot is a 3 point attempt, 2 otherwise
        """
        return 3 if "3pt Shot" in self.description else 2



"""
``StatsNbaBoxscoreLoader`` loads boxscore data for a game and
creates :obj:`~pbpstats.resources.boxscore.stats_nba_boxscore_item.StatsNbaBoxscoreItem`
objects for each player and team

The following code will load boxscore data for game id "0021900001" from
a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaBoxscoreLoader

    boxscore_loader = StatsNbaBoxscoreLoader("0021900001", "file", "/data")
    print(boxscore_loader.items[0].data) # prints dict with a player's boxscore data for game
"""
import json
import os

from pbpstats import NBA_STRING, G_LEAGUE_STRING
from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader
from pbpstats.resources.boxscore.stats_nba_boxscore_item import StatsNbaBoxscoreItem


class StatsNbaBoxscoreLoader(StatsNbaFileLoader, StatsNbaWebLoader):
    """
    Loads stats.nba.com source boxscore data for game.
    Team/Player data is stored in items attribute as :obj:`~pbpstats.resources.boxscore.stats_nba_boxscore_item.StatsNbaBoxscoreItem` objects

    :param str game_id: NBA Stats Game Id
    :param str source: Where should data be loaded from. Options are 'web' or 'file'
    :param str file_directory: (optional if source is 'web')
        Directory in which data should be either stored (if source is web) or loaded from (if source is file).
        The specific file location will be `stats_boxscore_<game_id>.json` in the `/game_details` subdirectory.
        If not provided response data will not be saved on disk.
    """

    data_provider = "stats_nba"
    resource = "Boxscore"
    parent_object = "Game"

    def __init__(self, game_id, source, file_directory=None):
        self.game_id = game_id
        self.file_directory = file_directory
        self.source = source
        self._load_data()
        self._make_boxscore_items()

    def _load_data(self):
        source_method = getattr(self, f"_from_{self.source}")
        source_method()

    @check_file_directory
    def _from_file(self):
        self.file_path = (
            f"{self.file_directory}/game_details/stats_boxscore_{self.game_id}.json"
        )
        self._load_data_from_file()

    def _from_web(self):
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league == G_LEAGUE_STRING
            else self.league
        )
        self.base_url = (
            f"https://stats.{league_url_part}.com/stats/boxscoretraditionalv2"
        )
        self.parameters = {
            "GameId": self.game_id,
            "StartPeriod": 0,
            "EndPeriod": 10,
            "RangeType": 2,
            "StartRange": 0,
            "EndRange": 55800,
        }
        self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = (
                f"{self.file_directory}/game_details/stats_boxscore_{self.game_id}.json"
            )
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    def _make_boxscore_items(self):
        self.items = [StatsNbaBoxscoreItem(item) for item in self.data]
        self.items += [
            StatsNbaBoxscoreItem(item)
            for item in self.make_list_of_dicts(results_set_index=1)
        ]
